import sys
import os

import pytest

from pubtools.pulplib import FakeController, RpmUnit, YumRepository, RpmDependency


@pytest.mark.parametrize("use_file_object", [False, True])
def test_can_upload_units(data_path, use_file_object):
    """repo.upload_rpm() succeeds with fake client and populates units."""

    rpm_path = os.path.join(data_path, "rpms/walrus-5.21-1.noarch.rpm")
    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    to_upload = rpm_path
    if use_file_object:
        to_upload = open(rpm_path, "rb")

    upload_f = repo1.upload_rpm(to_upload)

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = sorted(repo1.search_content().result(), key=lambda u: u.sha256sum)
    units_all = sorted(client.search_content().result(), key=lambda u: u.sha256sum)

    # They should be equal
    assert units_all == units_in_repo

    # And they should be this
    assert units_in_repo == [
        RpmUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="walrus",
            version="5.21",
            release="1",
            arch="noarch",
            epoch="0",
            signing_key="f78fb195",
            filename="walrus-5.21-1.noarch.rpm",
            sourcerpm="walrus-5.21-1.src.rpm",
            md5sum="6a3eec6d45e0ea80eab05870bf7a8d4b",
            sha1sum="8dea2b64fc52062d79d5f96ba6415bffae4d2153",
            sha256sum="e837a635cc99f967a70f34b268baa52e0f412c1502e08e924ff5b09f1f9573f2",
            content_type_id="rpm",
            repository_memberships=["repo1"],
            requires=[
                RpmDependency(
                    name="rpmlib(CompressedFileNames)",
                    version="3.0.4",
                    release="1",
                    flags="LE",
                    epoch="0",
                ),
                RpmDependency(
                    name="rpmlib(PayloadFilesHavePrefix)",
                    version="4.0",
                    release="1",
                    flags="LE",
                    epoch="0",
                ),
            ],
            provides=[
                RpmDependency(
                    name="walrus", version="5.21", release="1", flags="EQ", epoch="0"
                )
            ],
        )
    ]



import sys
import os

import pytest

from pubtools.pulplib import FakeController, ErratumUnit, YumRepository


def test_can_upload_units():
    """repo.upload_erratum() succeeds with fake client and populates units."""

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    to_upload = ErratumUnit(
        id="RHBA-1234:56",
        summary="test advisory",
        # Add some existing memberships in an attempt to screw it up...
        # (they should not affect the upload)
        repository_memberships=["repo2", "repo3"],
    )

    upload_f = repo1.upload_erratum(to_upload)

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = list(repo1.search_content())
    units_all = list(client.search_content())

    # They should be equal
    assert units_all == units_in_repo

    # And they should be this
    assert units_in_repo == [
        ErratumUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            id="RHBA-1234:56",
            summary="test advisory",
            repository_memberships=["repo1"],
        )
    ]


def test_upload_overwrite():
    """repo.upload_erratum() can overwrite fields of an existing advisory."""

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))
    controller.insert_repository(YumRepository(id="repo2"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()
    repo2 = client.get_repository("repo2").result()

    to_upload1 = ErratumUnit(id="RHBA-1234:56", summary="test advisory", version="1")
    to_upload2 = ErratumUnit(
        id="RHBA-1234:56",
        summary="updated test advisory",
        description="I've altered the deal",
        version="2",
    )
    to_upload3 = ErratumUnit(id="RHBA-1234:57", summary="a different advisory")

    # Upload all three of the above advisories.
    # Uploads 1 and 2 are for the same advisory (though we use
    # different repos to see what happens).
    assert repo1.upload_erratum(to_upload1).result()
    assert repo2.upload_erratum(to_upload2).result()
    assert repo2.upload_erratum(to_upload3).result()

    # Now let's check the outcome in each repo (and entire system).
    units_repo1 = sorted(repo1.search_content(), key=repr)
    units_repo2 = sorted(repo2.search_content(), key=repr)
    units_all = sorted(client.search_content(), key=repr)

    # What we expect to see is that the three uploads resulted in only
    # two errata, with the fields on the first advisory set to the values
    # after the most recent upload of it, and the second advisory equal to
    # the single upload for that advisory. Also, the first advisory is
    # now present in two repos.

    assert units_repo1 == [
        ErratumUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            id="RHBA-1234:56",
            summary="updated test advisory",
            description="I've altered the deal",
            version="2",
            repository_memberships=["repo1", "repo2"],
        )
    ]
    assert units_repo2 == [
        ErratumUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            id="RHBA-1234:56",
            summary="updated test advisory",
            description="I've altered the deal",
            version="2",
            repository_memberships=["repo1", "repo2"],
        ),
        ErratumUnit(
            unit_id="e6f4590b-9a16-4106-cf6a-659eb4862b21",
            id="RHBA-1234:57",
            summary="a different advisory",
            repository_memberships=["repo2"],
        ),
    ]

    assert units_repo2 == units_all


def test_upload_overwrite_noop():
    """repo.upload_erratum() doesn't overwrite erratum unit if version is unmodified."""

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))
    controller.insert_repository(YumRepository(id="repo2"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()
    repo2 = client.get_repository("repo2").result()

    to_upload1 = ErratumUnit(id="RHBA-1234:56", summary="test advisory", version="3")
    to_upload2 = ErratumUnit(
        id="RHBA-1234:56",
        summary="updated test advisory",
        description="changed it, but no effect due to same version",
        version="3",
    )

    # Upload the advisory to two different repos.
    assert repo1.upload_erratum(to_upload1).result()
    assert repo2.upload_erratum(to_upload2).result()

    # Now let's check the outcome in each repo (and entire system).
    units_repo1 = sorted(repo1.search_content(), key=repr)
    units_repo2 = sorted(repo2.search_content(), key=repr)
    units_all = sorted(client.search_content(), key=repr)

    # What we expect to see is that we only have one advisory (as only
    # a single id was used), that advisory is present in two repos, and
    # the advisory content is equal to the first upload because version
    # was not bumped for the second upload.

    assert units_repo1 == [
        ErratumUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            id="RHBA-1234:56",
            summary="test advisory",
            version="3",
            repository_memberships=["repo1", "repo2"],
        )
    ]
    assert units_repo2 == units_repo1
    assert units_all == units_repo1



import logging
import pytest

from pubtools.pulplib import (
    Repository,
    YumRepository,
    Task,
    Distributor,
    DetachedException,
    YumSyncOptions,
    TaskFailedException,
)


@pytest.fixture
def fixture_sync_async_response(requests_mocker):
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/sync/",
        [{"json": {"spawned_tasks": [{"task_id": "task1"}]}}],
    )


@pytest.fixture
def fixture_search_task_response(requests_mocker):
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [{"json": [{"task_id": "task1", "state": "finished"}]}],
    )


def test_detached():
    """sync raises if called on a detached repo"""
    with pytest.raises(DetachedException):
        Repository(id="some-repo").sync()


def test_sync_no_feed(
    fast_poller,
    client,
    requests_mocker,
    fixture_sync_async_response,
    fixture_search_task_response,
):
    """Test sync fail as no feed is provided."""
    repo = YumRepository(id="some-repo")
    repo.__dict__["_client"] = client

    # empty options should fail as feed is required to be non-empty
    try:
        repo.sync().result()
        assert "Exception should have been raised"
    except ValueError:
        pass


def test_sync_with_options(
    requests_mocker, client, fixture_sync_async_response, fixture_search_task_response
):
    """Test sync passes, test whether sync options are passed to override config."""
    repo = YumRepository(id="some-repo")
    repo.__dict__["_client"] = client

    options = YumSyncOptions(ssl_validation=False, feed="mock://example.com/")

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert repo.sync(options).result() == [
        Task(id="task1", succeeded=True, completed=True)
    ]

    req = requests_mocker.request_history

    assert req[0].json()["override_config"] == {
        "ssl_validation": False,
        "feed": "mock://example.com/",
    }



from pubtools.pulplib import (
    FakeController,
    YumRepository,
    YumSyncOptions,
    PulpException,
)


def test_can_sync():
    """repo.sync() succeeds with fake client and populates sync_history."""
    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))
    controller.insert_repository(YumRepository(id="repo2"))

    client = controller.client
    repo1 = client.get_repository("repo1")

    # Call to sync should succeed
    sync_f = repo1.sync(YumSyncOptions(feed="mock://feed/"))

    # The future should resolve successfully
    tasks = sync_f.result()

    # It should have returned at least one successful task.
    assert tasks
    for task in tasks:
        assert task.succeeded

    # The change should be reflected in the controller's sync history
    history = controller.sync_history

    assert len(history) == 1
    assert history[0].repository.id == "repo1"
    assert history[0].tasks == tasks


def test_sync_absent_raises():
    """repo.sync() of a nonexistent repo raises."""
    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo_copy1 = client.get_repository("repo1")
    repo_copy2 = client.get_repository("repo1")

    # If I delete the repo through one handle...
    assert repo_copy1.delete().result()

    # ...then sync through the other handle becomes impossible
    exception = repo_copy2.sync(YumSyncOptions(feed="mock://feed/")).exception()
    assert isinstance(exception, PulpException)
    assert "repo1 not found" in str(exception)



import logging
import pytest

from pubtools.pulplib import (
    Repository,
    YumRepository,
    Task,
    Distributor,
    DetachedException,
    PublishOptions,
    TaskFailedException,
)

from pubtools.pluggy import pm


@pytest.fixture
def hookspy():
    hooks = []

    def record_hook(hook_name, _hook_impls, kwargs):
        hooks.append((hook_name, kwargs))

    def do_nothing(*args, **kwargs):
        pass

    undo = pm.add_hookcall_monitoring(before=record_hook, after=do_nothing)
    yield hooks
    undo()


def test_detached():
    """publish raises if called on a detached repo"""
    with pytest.raises(DetachedException):
        Repository(id="some-repo").publish()


def test_publish_no_distributors(client):
    """publish succeeds and returns no tasks if repo contains no distributors."""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    assert repo.publish().result() == []


def test_publish_distributors(fast_poller, requests_mocker, client, hookspy):
    """publish succeeds and returns tasks from each applicable distributor"""
    repo = YumRepository(
        id="some-repo",
        distributors=(
            Distributor(id="yum_distributor", type_id="yum_distributor"),
            Distributor(id="other_distributor", type_id="other_distributor"),
            Distributor(id="cdn_distributor", type_id="rpm_rsync_distributor"),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}, {"task_id": "task2"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task3"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {
                "json": [
                    {"task_id": "task1", "state": "finished"},
                    {"task_id": "task2", "state": "skipped"},
                ]
            },
            {"json": [{"task_id": "task3", "state": "finished"}]},
        ],
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.publish()) == [
        Task(id="task1", succeeded=True, completed=True),
        Task(id="task2", succeeded=True, completed=True),
        Task(id="task3", succeeded=True, completed=True),
    ]

    # It should have first issued a request to publish yum_distributor
    req = requests_mocker.request_history
    assert (
        req[0].url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/"
    )
    assert req[0].json() == {"id": "yum_distributor", "override_config": {}}

    # Then polled for resulting tasks to succeed
    assert req[1].url == "https://pulp.example.com/pulp/api/v2/tasks/search/"
    assert req[1].json() == {
        "criteria": {"filters": {"task_id": {"$in": ["task1", "task2"]}}}
    }

    # Then published the next distributor
    assert (
        req[2].url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/"
    )
    assert req[2].json() == {"id": "cdn_distributor", "override_config": {}}

    # Then waited for those tasks to finish too
    assert req[3].url == "https://pulp.example.com/pulp/api/v2/tasks/search/"
    assert req[3].json() == {"criteria": {"filters": {"task_id": {"$in": ["task3"]}}}}

    # And there should have been no more requests
    assert len(req) == 4

    # It should have invoked hooks
    assert len(hookspy) == 2
    (hook_name, hook_kwargs) = hookspy[0]
    assert hook_name == "pulp_repository_pre_publish"
    assert hook_kwargs["repository"] is repo
    assert hook_kwargs["options"] == PublishOptions()
    (hook_name, hook_kwargs) = hookspy[1]
    assert hook_name == "pulp_repository_published"
    assert hook_kwargs["repository"] is repo
    assert hook_kwargs["options"] == PublishOptions()


def test_publish_with_options(requests_mocker, client):
    """publish passes expected config into distributors based on publish options"""
    repo = YumRepository(
        id="some-repo",
        distributors=(
            Distributor(id="yum_distributor", type_id="yum_distributor"),
            Distributor(id="cdn_distributor", type_id="rpm_rsync_distributor"),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task2"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
            {"json": [{"task_id": "task2", "state": "finished"}]},
        ],
    )

    options = PublishOptions(
        clean=True, force=True, origin_only=True, rsync_extra_args=["-a"]
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.publish(options)) == [
        Task(id="task1", succeeded=True, completed=True),
        Task(id="task2", succeeded=True, completed=True),
    ]

    req = requests_mocker.request_history

    # The yum_distributor request should have set force_full, but not
    # delete since it's not recognized by that distributor
    assert req[0].json()["override_config"] == {"force_full": True}

    # The cdn_distributor request should have set force_full, delete
    # and content_units_only
    assert req[2].json()["override_config"] == {
        "force_full": True,
        "delete": True,
        "content_units_only": True,
        "rsync_extra_args": ["-a"],
    }


def test_publish_fail(fast_poller, requests_mocker, client):
    """publish raises TaskFailedException if publish task fails"""
    repo = YumRepository(
        id="some-repo",
        distributors=(Distributor(id="yum_distributor", type_id="yum_distributor"),),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        json={"spawned_tasks": [{"task_id": "task1"}]},
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[
            {
                "task_id": "task1",
                "state": "error",
                "error": {"code": "ABC00123", "description": "Simulated error"},
            }
        ],
    )

    publish_f = repo.publish()

    # It should raise this exception
    with pytest.raises(TaskFailedException) as error:
        publish_f.result()

    # The exception should have a reference to the task which failed
    assert error.value.task.id == "task1"
    assert "Pulp task [task1] failed: ABC00123: Simulated error" in str(error.value)


def test_publish_broken_response(fast_poller, requests_mocker, client):
    """publish raises an exception if Pulp /publish/ responded with parseable
    JSON, but not of the expected structure
    """

    repo = YumRepository(
        id="some-repo",
        distributors=(Distributor(id="yum_distributor", type_id="yum_distributor"),),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        json={"spawned_tasks": ["oops, not a valid response"]},
    )

    publish_f = repo.publish()

    # It should raise some kind of exception due to the invalid spawned_tasks structure
    assert publish_f.exception()


def test_publish_retries(fast_poller, requests_mocker, client, caplog):
    """publish retries distributors as they fail"""
    caplog.set_level(logging.WARNING)

    repo = YumRepository(
        id="some-repo",
        distributors=(
            Distributor(id="yum_distributor", type_id="yum_distributor"),
            Distributor(id="cdn_distributor", type_id="cdn_distributor"),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task2"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task3"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
            {"json": [{"task_id": "task2", "state": "error"}]},
            {"json": [{"task_id": "task3", "state": "finished"}]},
        ],
    )

    publish_f = repo.publish()

    # It should succeed
    tasks = publish_f.result()

    # It should return only the *successful* tasks
    assert sorted([t.id for t in tasks]) == ["task1", "task3"]

    # Pick out the HTTP requests triggering distributors
    publish_reqs = [
        req
        for req in requests_mocker.request_history
        if req.url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/"
    ]
    publish_distributors = [req.json()["id"] for req in publish_reqs]

    # It should have triggered cdn_distributor twice, since the first attempt failed
    assert publish_distributors == [
        "yum_distributor",
        "cdn_distributor",
        "cdn_distributor",
    ]

    # The retry should have been logged
    messages = caplog.messages
    assert (
        messages[-1].splitlines()[0] == "Retrying due to error: Task task2 failed [1/6]"
    )



import datetime

import pytest

from pubtools.pulplib import (
    FakeController,
    Repository,
    Criteria,
    Matcher,
    Distributor,
    Task,
)


def test_can_search_id():
    """Can search for a repo by ID with fake client."""
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)

    client = controller.client
    crit = Criteria.with_id("repo1")
    found = client.search_repository(crit).data

    assert found == [repo1]


def test_can_search_ids():
    """Can search for a repo by list of IDs with fake client."""
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")
    repo3 = Repository(id="repo3")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_id(["repo1", "repo3", "other-id"])
    found = client.search_repository(crit).data

    assert sorted(found) == [repo1, repo3]


def test_can_search_id_exists():
    """Can search for a repo using exists operator with fake client."""
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)

    client = controller.client
    crit = Criteria.with_field("id", Matcher.exists())
    found = client.search_repository(crit).data

    assert sorted(found) == [repo1, repo2]


def test_search_no_result():
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)

    client = controller.client
    crit = Criteria.with_field("notes.whatever", "foobar")
    found = client.search_repository(crit).data

    assert found == []


def test_search_or():
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2")
    repo3 = Repository(id="repo3")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.or_(
        Criteria.with_id("repo3"), Criteria.with_field("id", Matcher.equals("repo1"))
    )
    found = client.search_repository(crit).data

    assert sorted(found) == [repo1, repo3]


def test_search_created_exists():
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2", created=datetime.datetime.utcnow())
    repo3 = Repository(id="repo3", created=datetime.datetime.utcnow())

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_field("notes.created", Matcher.exists())
    found = client.search_repository(crit).data

    assert sorted(found) == [repo2, repo3]


def test_search_with_model_field():
    """search repos using model field 'created' in criteria.
    returns only those repos which has the field and value is
    not None
    """
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2", created=datetime.datetime.utcnow())
    repo3 = Repository(id="repo3", created=None)

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_field("created", Matcher.exists())
    found = client.search_repository(crit).result().data

    assert sorted(found) == [repo2]


def test_search_and():
    controller = FakeController()

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2", created=datetime.datetime.utcnow())
    repo3 = Repository(id="repo3", created=datetime.datetime.utcnow())

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.and_(
        Criteria.with_field("notes.created", Criteria.exists), Criteria.with_id("repo2")
    )
    found = client.search_repository(crit).data

    assert sorted(found) == [repo2]


def test_search_null_and():
    """Search with an empty AND gives an error."""
    controller = FakeController()

    dist1 = Distributor(
        id="yum_distributor", type_id="yum_distributor", repo_id="repo1"
    )
    repo1 = Repository(id="repo1", distributors=[dist1])

    controller.insert_repository(repo1)
    controller.insert_task(Task(id="abc123"))

    client = controller.client
    crit = Criteria.and_()
    assert "Invalid AND in search query" in str(
        client.search_repository(crit).exception()
    )
    assert "Invalid AND in search query" in str(
        client.search_distributor(crit).exception()
    )
    assert "Invalid AND in search query" in str(client.search_task(crit).exception())


def test_search_null_or():
    """Search with an empty OR gives an error."""
    controller = FakeController()

    repo1 = Repository(id="repo1")

    controller.insert_repository(repo1)

    client = controller.client
    crit = Criteria.or_()
    assert "Invalid OR in search query" in str(
        client.search_repository(crit).exception()
    )


def test_search_bad_criteria():
    """Search with criteria of wrong type gives an error."""
    controller = FakeController()

    repo1 = Repository(id="repo1")

    controller.insert_repository(repo1)

    client = controller.client

    with pytest.raises(Exception) as repo_exc:
        client.search_repository("not a valid criteria")

    with pytest.raises(Exception) as dist_exc:
        client.search_distributor("invalid criteria")

    assert "Not a criteria" in str(repo_exc.value)
    assert "Not a criteria" in str(dist_exc.value)


def test_search_created_timestamp():
    controller = FakeController()

    when = datetime.datetime(2019, 6, 11, 14, 47, 0, tzinfo=None)
    when_str = "2019-06-11T14:47:00Z"

    repo1 = Repository(id="repo1")
    repo2 = Repository(id="repo2", created=when)
    repo3 = Repository(id="repo3", created=datetime.datetime.utcnow())

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_field("notes.created", when_str)
    found = client.search_repository(crit).data

    assert sorted(found) == [repo2]


def test_search_mapped_field_eq():
    """Can do equality search with fields subject to Python<=>Pulp conversion."""
    controller = FakeController()

    repo1 = Repository(id="repo1", eng_product_id=888)
    repo2 = Repository(id="repo2", signing_keys=["foo", "bar"])
    repo3 = Repository(id="repo3", eng_product_id=123)

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    keys_crit = Criteria.with_field("signing_keys", ["foo", "bar"])
    product_crit = Criteria.with_field("eng_product_id", 123)
    found_by_keys = client.search_repository(keys_crit).data
    found_by_product = client.search_repository(product_crit).data

    assert found_by_keys == [repo2]
    assert found_by_product == [repo3]


def test_search_mapped_field_in():
    """Can do 'in' search with fields subject to Python<=>Pulp conversion."""
    controller = FakeController()

    repo1 = Repository(id="repo1", eng_product_id=888)
    repo2 = Repository(id="repo2", eng_product_id=123)
    repo3 = Repository(id="repo3", eng_product_id=456)

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_field("eng_product_id", Matcher.in_([123, 456]))
    found = client.search_repository(crit).data

    assert sorted(found) == [repo2, repo3]


def test_search_mapped_field_regex():
    """Can do regex search with fields subject to Python<=>Pulp conversion."""
    controller = FakeController()

    repo1 = Repository(id="repo1", type="foobar")
    repo2 = Repository(id="repo2", type="foobaz")
    repo3 = Repository(id="repo3", type="quux")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)

    client = controller.client
    crit = Criteria.with_field("type", Matcher.regex("fooba[rz]"))
    found = client.search_repository(crit).data

    assert sorted(found) == [repo1, repo2]


def test_search_created_regex():
    """Can search using regular expressions."""

    controller = FakeController()

    when1 = datetime.datetime(2019, 6, 11, 14, 47, 0, tzinfo=None)
    when2 = datetime.datetime(2019, 3, 1, 1, 1, 0, tzinfo=None)
    when3 = datetime.datetime(2019, 6, 1, 1, 1, 0, tzinfo=None)

    repo1 = Repository(id="repo1", created=when1)
    repo2 = Repository(id="repo2", created=when2)
    repo3 = Repository(id="repo3", created=when3)
    repo4 = Repository(id="repo4")

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)
    controller.insert_repository(repo3)
    controller.insert_repository(repo4)

    client = controller.client
    crit = Criteria.with_field("notes.created", Matcher.regex("19-06"))
    found = client.search_repository(crit)

    assert sorted(found) == [repo1, repo3]


def test_search_paginates():
    controller = FakeController()

    repos = []
    for i in range(0, 1000):
        repo = Repository(id="repo-%s" % i)
        repos.append(repo)
        controller.insert_repository(repo)

    client = controller.client
    crit = Criteria.true()

    page = client.search_repository(crit)
    found_repos = list(page)

    page_count = 1
    while page.next:
        page_count += 1
        page = page.next.result()

    # There should have been several pages (it is not defined exactly
    # what page size the fake client uses, but it should be relatively
    # small to enforce that clients think about pagination)
    assert page_count >= 10

    # All repos should have been found
    assert sorted(found_repos) == sorted(repos)


def test_search_distributor():
    controller = FakeController()

    dist1 = Distributor(
        id="yum_distributor", type_id="yum_distributor", repo_id="repo1"
    )
    dist2 = Distributor(
        id="cdn_distributor", type_id="rpm_rsync_distributor", repo_id="repo1"
    )
    repo1 = Repository(id="repo1", distributors=(dist1, dist2))

    controller.insert_repository(repo1)

    client = controller.client
    crit = Criteria.true()

    found = client.search_distributor(crit).result().data

    assert sorted(found) == [dist2, dist1]


def test_search_distributor_with_relative_url():
    controller = FakeController()

    dist1 = Distributor(
        id="yum_distributor",
        type_id="yum_distributor",
        repo_id="repo1",
        relative_url="relative/path",
    )
    dist2 = Distributor(
        id="cdn_distributor",
        type_id="rpm_rsync_distributor",
        repo_id="repo1",
        relative_url="relative/path",
    )
    repo1 = Repository(id="repo1", distributors=(dist1, dist2))

    dist3 = Distributor(
        id="yum_distributor",
        type_id="yum_distributor",
        repo_id="repo2",
        relative_url="another/path",
    )

    repo2 = Repository(id="repo2", distributors=(dist3,))

    controller.insert_repository(repo1)
    controller.insert_repository(repo2)

    client = controller.client
    crit = Criteria.with_field("relative_url", Matcher.regex("relative/path"))

    found = client.search_distributor(crit).result().data

    assert sorted(found) == [dist2, dist1]


def test_search_mapped_field_less_than():
    controller = FakeController()

    dist1 = Distributor(
        id="yum_distributor",
        type_id="yum_distributor",
        repo_id="repo1",
        last_publish=datetime.datetime(2019, 8, 23, 2, 5, 0, tzinfo=None),
    )
    dist2 = Distributor(
        id="cdn_distributor",
        type_id="rpm_rsync_distributor",
        repo_id="repo1",
        last_publish=datetime.datetime(2019, 8, 27, 2, 5, 0, tzinfo=None),
    )
    repo1 = Repository(id="repo1", distributors=(dist1, dist2))

    controller.insert_repository(repo1)

    client = controller.client
    crit = Criteria.with_field(
        "last_publish", Matcher.less_than(datetime.datetime(2019, 8, 24, 0, 0, 0))
    )
    found = client.search_distributor(crit).result().data

    assert found == [dist1]


def test_search_task():
    controller = FakeController()

    task1 = Task(
        id="task1",
        completed=True,
        succeeded=True,
        tags=[
            "pulp:repository:repo1",
            "pulp:action:publish",
        ],
    )
    task2 = Task(
        id="task2",
        completed=True,
        succeeded=True,
        tags=[
            "pulp:repository:repo1",
            "pulp:action:import_upload",
        ],
    )

    controller.insert_task(task1)
    controller.insert_task(task2)

    # The tasks I inserted should be present
    assert controller.tasks == [task1, task2]

    client = controller.client

    crit = Criteria.with_field("tags", "pulp:action:publish")
    resp = client.search_task(crit).result().data
    assert resp == [task1]

    crit2 = Criteria.with_field("id", "task2")
    resp2 = client.search_task(crit2).result().data
    assert resp2 == [task2]



import datetime
import os

import pytest

from pubtools.pulplib import FakeController, FileRepository, YumRepository


def test_upload_file_meta_wrong_fields(tmpdir):
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    # This should immediately give an error
    with pytest.raises(ValueError) as excinfo:
        repo1.upload_file(
            str(somefile), description="My great file", size=48, other="whatever"
        )

    # It should give an indication of the problem
    assert "Not mutable FileUnit field(s): other, size" in str(excinfo.value)


def test_can_upload_file_meta(tmpdir):
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    upload_f = repo1.upload_file(
        str(somefile),
        description="My great file",
        cdn_path="/foo/bar.txt",
        version="2.0",
    )

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # File should now be in repo.
    units_in_repo = list(repo1.search_content())
    assert len(units_in_repo) == 1
    unit = units_in_repo[0]

    # Sanity check we got the right thing.
    assert unit.path == "some-file.txt"

    # Extra fields we passed during upload should be present here.
    assert unit.description == "My great file"
    assert unit.cdn_path == "/foo/bar.txt"
    assert unit.version == "2.0"


def test_can_reupload_file_meta(tmpdir):
    """Can overwrite by uploading same content twice with different metadata."""
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    time1 = datetime.datetime(2021, 12, 14, 14, 44, 0)
    time2 = datetime.datetime(2022, 1, 2, 3, 4, 5)

    upload_f = repo1.upload_file(
        str(somefile),
        description="My great file",
        cdn_path="/foo/bar.txt",
        cdn_published=time1,
    )

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # Now upload again, but with some different values.
    upload_f = repo1.upload_file(
        str(somefile), description="My even better file", cdn_published=time2
    )

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # File should now be in repo.
    units_in_repo = list(repo1.search_content())

    # Should result in just one unit, as the second upload effectively merges
    # with the existing unit.
    assert len(units_in_repo) == 1
    unit = units_in_repo[0]

    # Sanity check we got the right thing.
    assert unit.path == "some-file.txt"

    # Extra fields should be equal to the values passed in at the most
    # recent upload.
    assert unit.description == "My even better file"
    assert unit.cdn_published == time2

    # cdn_path was provided at the first upload and not the second.
    # In such cases it is expected that the field is wiped out, as you must
    # always provide *all* metadata at once.
    assert unit.cdn_path is None


def test_can_upload_rpm_meta(data_path):
    rpm_path = os.path.join(data_path, "rpms/walrus-5.21-1.noarch.rpm")
    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    to_upload = rpm_path

    upload_f = repo1.upload_rpm(to_upload, cdn_path="/path/to/my-great.rpm")

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # RPM should now be in repo.
    units_in_repo = list(repo1.search_content())
    assert len(units_in_repo) == 1
    unit = units_in_repo[0]

    # Sanity check we got the right unit.
    assert unit.filename == "walrus-5.21-1.noarch.rpm"

    # Extra fields we passed during upload should be present here.
    assert unit.cdn_path == "/path/to/my-great.rpm"



import pytest

from pubtools.pulplib._impl.fake.rpmlib import _parse_dep_relation


@pytest.mark.parametrize(
    "flag, expected_relation",
    [(0x00, ""), (0x02, "LT"), (0x04, "GT"), (0x08, "EQ"), (0x0A, "LE"), (0x0C, "GE")],
)
def test_parse_deps_relation(flag, expected_relation):
    """
    Tests expected value of relation according to input flags.
    """
    relation = _parse_dep_relation(flag)

    assert relation == expected_relation



import pytest

from pubtools.pulplib import (
    FakeController,
    Criteria,
    Matcher,
    RpmUnit,
    ErratumUnit,
    ModulemdUnit,
    YumRepository,
    RpmDependency,
)


@pytest.fixture
def controller():
    return FakeController()


@pytest.fixture
def populated_repo(controller):
    units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(
            content_type_id="srpm", name="bash", version="4.0", release="1", arch="src"
        ),
        RpmUnit(
            name="glibc",
            version="5.0",
            release="1",
            arch="x86_64",
            sourcerpm="glibc-5.0-1.el5_11.1.src.rpm",
            provides=[RpmDependency(name="gcc")],
        ),
        ModulemdUnit(
            name="module1", stream="s1", version=1234, context="a1b2", arch="x86_64"
        ),
        ModulemdUnit(
            name="module2", stream="s2", version=1234, context="a1b2", arch="x86_64"
        ),
        ErratumUnit(id="RHBA-1234:56", summary="The best advisory"),
    ]

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)
    controller.insert_units(repo, units)

    return controller.client.get_repository(repo.id).result()


def test_search_content_empty_repo(controller):
    """search_content on empty repo gives no results"""
    repo = YumRepository(id="empty-repo")
    controller.insert_repository(repo)

    assert list(controller.client.get_repository("empty-repo").search_content()) == []


def test_search_content_missing_repo(controller, populated_repo):
    """search_content on nonexistent repo raises"""
    # Get an additional reference to the same repo
    repo = controller.client.get_repository("repo1")

    # Now delete the repo through one reference
    populated_repo.delete().result()

    # Searching through the other reference should now return a failed future
    # with reasonable error
    assert "Repository id=repo1 not found" in str(repo.search_content().exception())


def test_search_content_unsupported_operator(populated_repo):
    """search_content using unsupported operators on content_type_id raises"""
    with pytest.raises(ValueError) as e:
        populated_repo.search_content(
            Criteria.with_field("content_type_id", Matcher.regex("foobar"))
        )

    assert "unsupported expression for content_type_id" in str(e.value)


def test_search_null_and(populated_repo):
    """Search with an empty AND gives an error."""
    crit = Criteria.and_()
    assert "Invalid AND in search query" in str(
        populated_repo.search_content(crit).exception()
    )


def test_search_content_default_crit(populated_repo):
    """search_content with default criteria on populated repo finds all units"""

    units = list(populated_repo.search_content())
    assert len(units) == 6


def test_search_content_by_type(populated_repo):
    """search_content for particular type returns matching content"""

    crit = Criteria.with_field("content_type_id", "rpm")
    units = list(populated_repo.search_content(crit))
    assert sorted(units) == [
        RpmUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        RpmUnit(
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
            name="glibc",
            version="5.0",
            release="1",
            arch="x86_64",
            sourcerpm="glibc-5.0-1.el5_11.1.src.rpm",
            repository_memberships=["repo1"],
            provides=[RpmDependency(name="gcc")],
        ),
    ]


def test_search_erratum_by_type(populated_repo):
    """search_content for erratum returns matching content"""

    crit = Criteria.with_field("content_type_id", "erratum")
    units = list(populated_repo.search_content(crit))
    assert units == [
        ErratumUnit(
            unit_id="85776e9a-dd84-f39e-7154-5a137a1d5006",
            id="RHBA-1234:56",
            summary="The best advisory",
            repository_memberships=["repo1"],
        )
    ]


def test_search_content_by_unit_field(populated_repo):
    """search_content on regular field returns matching content"""

    crit = Criteria.with_field("name", "bash")
    units = list(populated_repo.search_content(crit))
    assert sorted(units) == [
        RpmUnit(
            unit_id="82e2e662-f728-b4fa-4248-5e3a0a5d2f34",
            content_type_id="srpm",
            name="bash",
            version="4.0",
            release="1",
            arch="src",
            repository_memberships=["repo1"],
        ),
        RpmUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
    ]


def test_search_content_by_unit_type(populated_repo):
    """search_content on unit_type returns only units of that type"""

    crit = Criteria.with_unit_type(ModulemdUnit)
    units = list(populated_repo.search_content(crit))
    assert sorted(units) == [
        ModulemdUnit(
            unit_id="23a7711a-8133-2876-37eb-dcd9e87a1613",
            name="module1",
            stream="s1",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        ModulemdUnit(
            unit_id="e6f4590b-9a16-4106-cf6a-659eb4862b21",
            name="module2",
            stream="s2",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
    ]


def test_search_content_mixed_fields(populated_repo):
    """search_content crossing multiple fields and types returns matching units"""

    crit = Criteria.and_(
        Criteria.with_field_in("content_type_id", ["rpm", "modulemd"]),
        Criteria.with_field_in("name", ["bash", "module1"]),
    )
    units = list(populated_repo.search_content(crit))

    # Note: sorting different types not natively supported, hence sorting by repr
    assert sorted(units, key=repr) == [
        ModulemdUnit(
            unit_id="23a7711a-8133-2876-37eb-dcd9e87a1613",
            name="module1",
            stream="s1",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
        RpmUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            repository_memberships=["repo1"],
        ),
    ]


def test_search_content_subfields(populated_repo):
    """
    search_content using subfield in attributes that are lists of objects.
    This applies for 'provides' and 'requires' attributes in RpmUnit.
    """
    crit = Criteria.and_(
        Criteria.with_unit_type(RpmUnit), Criteria.with_field("provides.name", "gcc")
    )
    units = list(populated_repo.search_content(crit))

    assert units == [
        RpmUnit(
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
            name="glibc",
            version="5.0",
            release="1",
            arch="x86_64",
            sourcerpm="glibc-5.0-1.el5_11.1.src.rpm",
            provides=[RpmDependency(name="gcc")],
            repository_memberships=["repo1"],
        )
    ]



import os
from io import BytesIO

from xml.parsers.expat import ExpatError

import pytest

from pubtools.pulplib import FakeController, YumRepository


def test_can_upload_comps(data_path):
    """repo.upload_comps_xml() succeeds with fake client."""

    xml_path = os.path.join(data_path, "sample-comps.xml")

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    upload_f = repo1.upload_comps_xml(xml_path)

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # If I now search for all content...
    units_all = list(client.search_content())

    # There should still be nothing, as the fake does not actually store
    # and reproduce comps-related units.
    assert units_all == []


def test_upload_comps_error():
    """repo.upload_comps_xml() raises with fake client when given invalid input."""
    xml = BytesIO(b"Oops not valid")

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    with pytest.raises(ExpatError):
        repo1.upload_comps_xml(xml)



# -*- coding: utf-8 -*-
import logging
import time
import pytest
import json

from six.moves import StringIO

from pubtools.pulplib import (
    Repository,
    FileRepository,
    Task,
    DetachedException,
    TaskFailedException,
)

from pubtools.pulplib._impl.log import TimedLogger
from pubtools.pulplib._impl.client import client

from ..ioutil import ZeroesIO


@pytest.fixture()
def fast_timed_logger(monkeypatch):
    # Force TimedLogger class to use an interval of 0 (i.e. it logs every time).
    # Tests otherwise will run too fast for any of the log messages to ever
    # be produced.

    class FastTimedLogger(TimedLogger):
        def __init__(self):
            super(FastTimedLogger, self).__init__(interval=0)

    monkeypatch.setattr(client, "TimedLogger", FastTimedLogger)


def test_upload_detached():
    """upload_file raises if called on a detached repo"""
    with pytest.raises(DetachedException):
        FileRepository(id="some-repo").upload_file("some-file")


def test_upload_file(client, requests_mocker, tmpdir, caplog, fast_timed_logger):
    """test upload a file to a repo in pulp"""

    logging.getLogger().setLevel(logging.INFO)
    caplog.set_level(logging.INFO)

    repo_id = "repo1"
    repo = FileRepository(id=repo_id)
    repo.__dict__["_client"] = client

    client._CHUNK_SIZE = 20

    request_body = {
        "_href": "/pulp/api/v2/content/uploads/cfb1fed0-752b-439e-aa68-fba68eababa3/",
        "upload_id": "cfb1fed0-752b-439e-aa68-fba68eababa3",
    }
    upload_id = request_body["upload_id"]

    import_report = {
        "result": {},
        "error": {},
        "spawned_tasks": [{"_href": "/pulp/api/v2/tasks/task1/", "task_id": "task1"}],
    }

    tasks_report = [{"task_id": "task1", "state": "finished"}]

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/content/uploads/", json=request_body
    )
    requests_mocker.put(
        "https://pulp.example.com/pulp/api/v2/content/uploads/%s/0/" % upload_id,
        json=[],
    )
    requests_mocker.put(
        "https://pulp.example.com/pulp/api/v2/content/uploads/%s/20/" % upload_id,
        json=[],
    )
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/%s/actions/import_upload/"
        % repo_id,
        json=import_report,
    )
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/", json=tasks_report
    )
    requests_mocker.delete(
        "https://pulp.example.com/pulp/api/v2/content/uploads/%s/" % upload_id, json=[]
    )

    assert repo.upload_file(str(somefile)).result() == [
        Task(id="task1", succeeded=True, completed=True)
    ]

    # the 6th request call might not be done in time, try 1000
    # times with .01 sec sleep before next try.
    for i in range(1000):
        time.sleep(0.01)
        try:
            assert requests_mocker.call_count == 6
        except AssertionError:
            if i != 999:
                continue
            else:
                raise
        else:
            break

    # 4th call should be import, check if right unit_key's passed
    import_request = requests_mocker.request_history[3].json()
    import_unit_key = {
        u"name": somefile.basename,
        u"checksum": u"fad3fc1e6d583b2003ec0a5273702ed8fcc2504271c87c40d9176467ebe218cb",
        u"size": 29,
    }
    assert import_request["unit_key"] == import_unit_key

    messages = caplog.messages

    # It should tell us about the upload
    assert (
        "Uploading some-file.txt to repo1 [cfb1fed0-752b-439e-aa68-fba68eababa3]"
        in messages
    )

    # It should log some progress info during the upload.
    # (Note these logs would not normally be produced for such a small amount of
    # progress - it's only because we have reduced the logging interval for testing)
    assert (
        "Still uploading some-file.txt: 20 Bytes / 68% [cfb1fed0-752b-439e-aa68-fba68eababa3]"
        in messages
    )
    assert (
        "Still uploading some-file.txt: 29 Bytes / 100% [cfb1fed0-752b-439e-aa68-fba68eababa3]"
        in messages
    )

    # task's spawned and completed
    assert "Created Pulp task: task1" in messages
    assert "Pulp task completed: task1" in messages


@pytest.mark.parametrize(
    "relative_url,expected",
    [
        ("some/path/", "some/path/some-file.txt"),
        ("some/path/foo.txt", "some/path/foo.txt"),
    ],
)
def test_get_relative_url(tmpdir, relative_url, expected):
    somefile = tmpdir.join("some-file.txt")
    repo = FileRepository(id="some-repo")
    result = repo._get_relative_url(str(somefile), relative_url)

    assert result == expected


def test_get_relative_url_with_file_object(tmpdir):
    repo = FileRepository(id="some-repo")
    file_obj = StringIO()

    with pytest.raises(ValueError):
        repo._get_relative_url(file_obj, None)

    with pytest.raises(ValueError):
        repo._get_relative_url(file_obj, "some/path/")

    assert repo._get_relative_url(file_obj, "path/foo.txt") == "path/foo.txt"


def test_upload_file_contains_unicode(client, requests_mocker):
    file_obj = StringIO("哈罗")
    upload_id = "cfb1fed0-752b-439e-aa68-fba68eababa3"

    requests_mocker.put(
        "https://pulp.example.com/pulp/api/v2/content/uploads/%s/0/" % upload_id,
        json=[],
    )

    repo_id = "repo1"
    repo = FileRepository(id=repo_id)
    repo.__dict__["_client"] = client

    upload_f = client._do_upload_file(upload_id, file_obj)

    assert upload_f.result() == (
        "478f4808df7898528c7f13dc840aa321c4109f5c9f33bad7afcffc0253d4ff8f",
        6,
    )


def test_upload_file_verylarge(client, requests_mocker):
    """Client can upload a 2GB file successfully."""

    file_size = 2000000000
    file_obj = ZeroesIO(file_size)

    upload_id = "cfb1fed0-752b-439e-aa68-fba68eababa3"

    # We will be uploading many chunks and we need to mock the 'offset' for each one...
    # Note this must align with the client's internal CHUNK_SIZE.
    chunk_size = 1024 * 1024
    for i in range(0, 3000):
        requests_mocker.put(
            "https://pulp.example.com/pulp/api/v2/content/uploads/%s/%d/"
            % (upload_id, i * chunk_size),
            json=[],
        )

    repo_id = "repo1"
    repo = FileRepository(id=repo_id)
    repo.__dict__["_client"] = client

    # It should be able to upload successfully.
    upload_f = client._do_upload_file(upload_id, file_obj)

    assert upload_f.result() == (
        # We should get the right checksum and size back, which proves all
        # the data was read correctly.
        #
        # If you want to verify this checksum, try:
        #
        #   dd if=/dev/zero bs=1000000 count=2000 status=none | sha256sum
        #
        "2e0c654b6cba3a1e816726bae0eac481eb7fd0351633768c3c18392e0f02b619",
        file_size,
    )



# Provide a simple fallback for humanize.naturalsize in
# certain legacy environments.
#
# TODO: remove me once py2 support goes away.

# pylint: disable=broad-except


def fallback_naturalsize(value):
    return "%.1f MB" % (float(value) / 1024 / 1024)


try:
    from humanize import naturalsize
except Exception:  # pragma: no cover
    naturalsize = fallback_naturalsize

__all__ = ["naturalsize"]



import pytest

from pubtools.pulplib import Repository, YumRepository, DetachedException


def test_from_data_gives_yum_repository():
    """Repository.from_data maps to YumRepository subclass when needed"""
    repo = Repository.from_data({"id": "my-repo", "notes": {"_repo-type": "rpm-repo"}})
    assert isinstance(repo, YumRepository)


def test_default_mutable_urls():
    """mutable_urls has expected default value"""
    assert YumRepository(id="foo").mutable_urls == ["repodata/repomd.xml"]


def test_can_hash():
    """a default YumRepository is hashable"""
    repo = YumRepository(id="foo")
    reposet = set()
    reposet.add(repo)
    assert repo in reposet


def test_from_data_relative_url():
    """relative_url is initialized from distributors when possible"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "rpm-repo"},
            "distributors": [
                {
                    "id": "yum_distributor",
                    "distributor_type_id": "yum_distributor",
                    "config": {"relative_url": "some/publish/path"},
                }
            ],
        }
    )

    assert repo.relative_url == "some/publish/path"


def test_from_data_skip_rsync_repodata():
    """skip_rsync_repodata is initialized from distributors when possible"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "rpm-repo"},
            "distributors": [
                {
                    "id": "cdn_distributor",
                    "distributor_type_id": "rpm_rsync_distributor",
                    "config": {"skip_repodata": True},
                }
            ],
        }
    )
    assert repo.skip_rsync_repodata


def test_populate_attrs():
    """test populate attributes are correctly parsed from repo notes"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {
                "_repo-type": "rpm-repo",
                "content_set": "fake_content_set",
                "population_sources": ["populate_repo1", "populate_repo2"],
                "ubi_population": True,
                "ubi_config_version": "fake_ubi_config_version",
            },
            "distributors": [],
        }
    )
    assert repo.population_sources == ["populate_repo1", "populate_repo2"]
    assert repo.ubi_population
    assert repo.content_set == "fake_content_set"
    assert repo.ubi_config_version == "fake_ubi_config_version"


def test_related_repositories(client, requests_mocker):
    """test Repository.get_*_repository returns expected objects"""

    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/distributors/search/",
        [
            {
                "json": [
                    {
                        "id": "yum_distributor",
                        "distributor_type_id": "yum_distributor",
                        "repo_id": "repo_debug",
                        "config": {"relative_url": "some/repo/debug"},
                    }
                ]
            },
            {
                "json": [
                    {
                        "id": "yum_distributor",
                        "distributor_type_id": "yum_distributor",
                        "repo_id": "repo_source",
                        "config": {"relative_url": "some/repo/SRPMS"},
                    }
                ]
            },
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/search/",
        [{"json": [{"id": "repo_debug"}]}, {"json": [{"id": "repo_source"}]}],
    )

    # Request for binary repo should return identical object
    assert repo_binary_test is repo_binary_test.get_binary_repository().result()
    assert repo_binary_test.get_binary_repository().id == "repo_binary"
    # Requests for debug and source repositories return correct repositories
    assert repo_binary_test.get_debug_repository().id == "repo_debug"
    assert repo_binary_test.get_source_repository().id == "repo_source"


def test_related_repositories_not_found(client, requests_mocker):
    """test Repository.get_*_repository returns Future[None] if repository is not found"""

    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/distributors/search/", json=[]
    )

    repo = repo_binary_test.get_source_repository()
    assert repo.result() is None


def test_related_repositories_detached_client():
    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = None

    with pytest.raises(DetachedException):
        repo_binary_test.get_binary_repository()



from pubtools.pulplib import ContainerImageRepository, Distributor, PublishOptions, Task


def test_publish_order(requests_mocker, client):
    """publish runs docker/rsync distributors in correct order"""
    repo = ContainerImageRepository(
        id="some-repo",
        distributors=(
            Distributor(
                id="docker_web_distributor_name_cli", type_id="docker_distributor_web"
            ),
            Distributor(id="cdn_distributor", type_id="docker_rsync_distributor"),
            Distributor(
                id="cdn_distributor_unprotected", type_id="docker_rsync_distributor"
            ),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task2"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task3"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
            {"json": [{"task_id": "task2", "state": "finished"}]},
            {"json": [{"task_id": "task3", "state": "finished"}]},
        ],
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.publish()) == [
        Task(id="task1", succeeded=True, completed=True),
        Task(id="task2", succeeded=True, completed=True),
        Task(id="task3", succeeded=True, completed=True),
    ]

    req = requests_mocker.request_history
    ids = [r.json()["id"] for r in req if r.url.endswith("/publish/")]

    # It should have triggered these distributors in this order
    assert ids == [
        "cdn_distributor",
        "cdn_distributor_unprotected",
        "docker_web_distributor_name_cli",
    ]


def test_publish_origin_only(requests_mocker, client):
    """publish skips docker distributor if origin_only=True"""
    repo = ContainerImageRepository(
        id="some-repo",
        distributors=(
            Distributor(
                id="docker_web_distributor_name_cli", type_id="docker_distributor_web"
            ),
            Distributor(id="cdn_distributor", type_id="docker_rsync_distributor"),
            Distributor(
                id="cdn_distributor_unprotected", type_id="docker_rsync_distributor"
            ),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task2"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
            {"json": [{"task_id": "task2", "state": "finished"}]},
        ],
    )

    # It should have succeeded, with the tasks as retrieved from Pulp
    assert sorted(repo.publish(PublishOptions(origin_only=True))) == [
        Task(id="task1", succeeded=True, completed=True),
        Task(id="task2", succeeded=True, completed=True),
    ]

    req = requests_mocker.request_history
    publish_req = [r for r in req if r.url.endswith("/publish/")]

    # It should have triggered these distributors in this order
    ids = [r.json()["id"] for r in publish_req]
    assert ids == ["cdn_distributor", "cdn_distributor_unprotected"]

    # It should have triggered with content_units_only=True
    assert publish_req[-1].json()["override_config"]["content_units_only"]



from pubtools.pulplib import FakeController, YumRepository, Distributor, PulpException


def test_can_publish():
    """repo.publish() succeeds with fake client and populates publish_history."""
    controller = FakeController()

    controller.insert_repository(
        YumRepository(
            id="repo1",
            distributors=[
                Distributor(id="yum_distributor", type_id="yum_distributor"),
                Distributor(id="cdn_distributor", type_id="rpm_rsync_distributor"),
            ],
        )
    )
    controller.insert_repository(YumRepository(id="repo2"))

    client = controller.client
    repo1 = client.get_repository("repo1")

    # Call to publish should succeed
    publish_f = repo1.publish()

    # The future should resolve successfully
    tasks = publish_f.result()

    # It should have returned at least one successful task.
    assert tasks
    for task in tasks:
        assert task.succeeded

    # The change should be reflected in the controller's publish history
    history = controller.publish_history

    assert len(history) == 1
    assert history[0].repository.id == "repo1"
    assert history[0].tasks == tasks


def test_publish_absent_raises():
    """repo.publish() of a nonexistent repo raises."""
    controller = FakeController()

    controller.insert_repository(
        YumRepository(
            id="repo1",
            distributors=[
                Distributor(id="yum_distributor", type_id="yum_distributor"),
                Distributor(id="cdn_distributor", type_id="rpm_rsync_distributor"),
            ],
        )
    )

    client = controller.client
    repo_copy1 = client.get_repository("repo1")
    repo_copy2 = client.get_repository("repo1")

    # If I delete the repo through one handle...
    assert repo_copy1.delete().result()

    # ...then publish through the other handle becomes impossible
    exception = repo_copy2.publish().exception()
    assert isinstance(exception, PulpException)
    assert "repo1 not found" in str(exception)



from pubtools.pulplib import FakeController, Repository, Distributor


def test_can_delete():
    """distributor.delete() with fake client removes distributor from fake data."""
    controller = FakeController()

    repo1 = Repository(
        id="repo1",
        distributors=[
            Distributor(id="dist1", type_id="type1", repo_id="repo1"),
            Distributor(id="dist2", type_id="type2", repo_id="repo1"),
        ],
    )

    controller.insert_repository(repo1)

    client = controller.client
    repo1 = client.get_repository("repo1")

    # Call to delete should succeed
    delete_f = repo1.distributors[0].delete()

    # The future should resolve successfully
    tasks = delete_f.result()

    # It should have returned at least one successful task.
    assert tasks
    for task in tasks:
        assert task.succeeded

    # The change should be reflected in the controller, with the
    # deleted distributor no longer present
    assert controller.repositories == [
        Repository(
            id="repo1",
            distributors=[Distributor(id="dist2", type_id="type2", repo_id="repo1")],
        )
    ]


def test_delete_missing_repo_fails():
    """dist.delete() of distributor with absent repo fails."""
    controller = FakeController()

    controller.insert_repository(
        Repository(
            id="repo",
            distributors=[
                Distributor(id="somedist", type_id="sometype", repo_id="repo")
            ],
        )
    )

    client = controller.client

    # Get two handles to the repo, to set up the situation that
    # the distributor is not detached despite repo deletion
    repo1 = client.get_repository("repo")
    repo2 = client.get_repository("repo")
    dist = repo2.distributors[0]

    # We can delete the repo
    assert repo1.delete().result()

    # The distributor (obtained via repo2) is not detached, but nevertheless
    # deletion should fail as repo no longer exists
    assert "Repository id=repo not found" in str(dist.delete().exception())


def test_delete_missing_distributor_succeeds():
    """dist.delete() of absent distributor with fake client succeeds.

    Deleting a distributor succeeds with the fake client since it also succeeds
    with the real client.
    """
    controller = FakeController()

    controller.insert_repository(
        Repository(
            id="repo",
            distributors=[
                Distributor(id="somedist", type_id="sometype", repo_id="repo")
            ],
        )
    )

    client = controller.client

    # Get two handles to the repos/distributors so we can delete via one handle
    # and try to delete again via the other
    repo_copy1 = client.get_repository("repo")
    repo_copy2 = client.get_repository("repo")

    dist_copy1 = repo_copy1.distributors[0]
    dist_copy2 = repo_copy2.distributors[0]

    # First delete succeeds, with some tasks
    assert dist_copy1.delete().result()

    # Second delete also succeeds, but there are no tasks since distributor
    # already doesn't exist
    assert dist_copy2.delete().result() == []



import pytest

from pubtools.pulplib import (
    FakeController,
    Criteria,
    Matcher,
    RpmUnit,
    ErratumUnit,
    ModulemdUnit,
    YumRepository,
)


@pytest.fixture
def controller():
    return FakeController()


def test_copy_content_empty_repo(controller):
    """copy_content from empty repo succeeds and copies nothing"""
    src = YumRepository(id="empty-repo")
    dest = YumRepository(id="dest-repo")
    controller.insert_repository(src)
    controller.insert_repository(dest)

    client = controller.client

    # Repos are initially detached, re-fetch them via client
    src = client.get_repository(src.id).result()
    dest = client.get_repository(dest.id).result()

    # It should succeed
    copy_tasks = list(client.copy_content(src, dest))

    # It should have at least one task
    assert copy_tasks

    for t in copy_tasks:
        # Every task should have succeeded
        assert t.succeeded

        # No units should have been copied
        assert not t.units


def test_copy_content_all(controller):
    """copy_content with no criteria copies all units from a repo"""
    src = YumRepository(id="src-repo")
    dest = YumRepository(id="dest-repo")
    other = YumRepository(id="other-repo")
    controller.insert_repository(src)
    controller.insert_repository(dest)
    controller.insert_repository(other)

    # Set up that both 'src' and 'other' contain some units.
    # The units in 'other' are there to ensure that copying only
    # happens from the given source repo and not other repos unexpectedly.
    src_units = [
        ErratumUnit(id="RHSA-1111:22", summary="Fixes bad things"),
        ModulemdUnit(
            name="module1",
            stream="s1",
            version=1234,
            context="a1b2",
            arch="x86_64",
            repository_memberships=["repoA", "repoB"],
        ),
        RpmUnit(
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            # Note: the next two fields aren't part of the unit key on RPM units.
            # We put these here specifically to verify they'll be filtered out
            # on the copy response (as a real Pulp server does).
            filename="bash-4.0-1.x86_64.rpm",
            signing_key="a1b2c3",
        ),
    ]
    controller.insert_units(src, src_units)
    controller.insert_units(
        other,
        [
            RpmUnit(
                name="glibc",
                version="5.0",
                release="1",
                arch="x86_64",
                sourcerpm="glibc-5.0-1.el5_11.1.src.rpm",
            )
        ],
    )

    client = controller.client

    # Repos are initially detached, re-fetch them via client
    src = client.get_repository(src.id).result()
    dest = client.get_repository(dest.id).result()

    # It should succeed
    copy_tasks = list(client.copy_content(src, dest))

    # Gather all units apparently copied
    units = sum([t.units for t in copy_tasks], [])

    # It should copy just the units we expect, from src.
    # Note that these are incomplete views of the units, as (just like real pulp)
    # the fake will only return fields present in the unit_key after a copy.
    assert sorted(units, key=repr) == [
        ErratumUnit(id="RHSA-1111:22"),
        ModulemdUnit(
            name="module1", stream="s1", version=1234, context="a1b2", arch="x86_64"
        ),
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64", epoch="0"),
    ]

    # The copy should also impact subsequent content searches.
    dest_units = list(dest.search_content())

    # The units we get from the search are not *precisely* the same as src_units,
    # because repository_memberships has been updated.
    assert sorted(dest_units, key=repr) == [
        ErratumUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            id="RHSA-1111:22",
            summary="Fixes bad things",
            content_type_id="erratum",
            repository_memberships=["src-repo", "dest-repo"],
        ),
        ModulemdUnit(
            unit_id="82e2e662-f728-b4fa-4248-5e3a0a5d2f34",
            name="module1",
            stream="s1",
            version=1234,
            context="a1b2",
            arch="x86_64",
            content_type_id="modulemd",
            repository_memberships=["src-repo", "dest-repo", "repoA", "repoB"],
        ),
        RpmUnit(
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            epoch="0",
            signing_key="a1b2c3",
            filename="bash-4.0-1.x86_64.rpm",
            content_type_id="rpm",
            repository_memberships=["src-repo", "dest-repo"],
        ),
    ]


def test_copy_content_with_criteria(controller):
    """copy_content can filter copied units by field values"""

    src = YumRepository(id="src-repo")
    dest = YumRepository(id="dest-repo")
    controller.insert_repository(src)
    controller.insert_repository(dest)

    src_units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(name="bash", version="4.0", release="2", arch="x86_64"),
        RpmUnit(name="bash", version="4.1", release="3", arch="x86_64"),
        RpmUnit(name="glibc", version="5.0", release="1", arch="x86_64"),
    ]
    controller.insert_units(src, src_units)

    client = controller.client

    # Repos are initially detached, re-fetch them via client
    src = client.get_repository(src.id).result()
    dest = client.get_repository(dest.id).result()

    # This is what we want to copy...
    crit = Criteria.and_(
        Criteria.with_field("name", "bash"),
        Criteria.with_field("release", Matcher.in_(["1", "3"])),
    )

    # Copy should succeed
    copy_tasks = list(client.copy_content(src, dest, crit))

    # It should have copied only those units matching the criteria
    units = sum([t.units for t in copy_tasks], [])
    assert sorted(units, key=repr) == [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64", epoch="0"),
        RpmUnit(name="bash", version="4.1", release="3", arch="x86_64", epoch="0"),
    ]

    # The copy should also impact subsequent content searches.
    dest_units = list(dest.search_content())
    assert sorted(dest_units, key=repr) == [
        RpmUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            epoch="0",
            repository_memberships=["src-repo", "dest-repo"],
        ),
        RpmUnit(
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
            name="bash",
            version="4.1",
            release="3",
            arch="x86_64",
            epoch="0",
            repository_memberships=["src-repo", "dest-repo"],
        ),
    ]



from pubtools.pulplib import FakeController, Repository


def test_can_delete():
    """repo.delete() with fake client removes repository from fake data."""
    controller = FakeController()

    controller.insert_repository(Repository(id="repo1"))
    controller.insert_repository(Repository(id="repo2"))

    client = controller.client
    repo1 = client.get_repository("repo1")

    # Call to delete should succeed
    delete_f = repo1.delete()

    # The future should resolve successfully
    tasks = delete_f.result()

    # It should have returned at least one successful task.
    assert tasks
    for task in tasks:
        assert task.succeeded

    # The change should be reflected in the controller, with the
    # deleted repo no longer present
    assert controller.repositories == [Repository(id="repo2")]


def test_delete_missing_repo_succeeds():
    """repo.delete() of absent repo with fake client succeeds.

    Deleting a repo succeeds with the fake client since it also succeeds
    with the real client.
    """
    controller = FakeController()

    controller.insert_repository(Repository(id="repo"))

    client = controller.client

    # We want to try deleting the same repo more than once.
    # We can't do this through a single reference since it would be detached
    # on delete, but if we get two handles to the same repo, we can.
    repo_copy1 = client.get_repository("repo")
    repo_copy2 = client.get_repository("repo")

    # First delete succeeds, with some tasks
    assert repo_copy1.delete().result()

    # Second delete also succeeds, but there are no tasks since repo
    # already doesn't exist
    assert repo_copy2.delete().result() == []
